﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.Windows.Media; // Geometry
using System.Windows; // Rect
using System.Windows.Input; //Key
using System.Windows.Threading; // DispatcherTimer

namespace WpfApp1_flappy
{
    abstract class GameItem
    {
        protected Geometry area; // (0;0) centered!
        protected double rotDegree = 0; // rotation in degrees

        public double CX { get; set; } // center position when rendering
        public double CY { get; set; }

        public double Rad
        {
            get
            {
                return rotDegree * Math.PI / 180;
            }
            set
            {
                rotDegree = 180 * value / Math.PI;
            }
        }

        public Geometry RealArea
        {
            get
            {
                TransformGroup tg = new TransformGroup();
                tg.Children.Add(new TranslateTransform(CX, CY));
                tg.Children.Add(new RotateTransform(rotDegree, CX, CY));
                // Geometry copy = area.Clone();
                area.Transform = tg;
                return area.GetFlattenedPathGeometry();
            }
        }

        public bool IsCollision(GameItem other)
        {
            return Geometry.Combine(RealArea, other.RealArea,
                GeometryCombineMode.Intersect, null).GetArea() > 0;
        }
    }

    class OneBird : GameItem
    {
        public double DY { get; set; }
        public double DX { get; set; }

        public OneBird(double newcx, double newcy)
        {
            CX = newcx;
            CY = newcy;
            DX = 10;

            Point p1 = new Point(10, 0);
            Point p2 = new Point(-10, -10);
            Point p3 = new Point(-10, 10);

            GeometryGroup g = new GeometryGroup();
            g.Children.Add(new LineGeometry(p1, p2));
            g.Children.Add(new LineGeometry(p1, p3));
            area = g.GetWidenedPathGeometry(new Pen(Brushes.Black, 2));
        }
    }

    class OnePipe : GameItem
    {
        public const int Gap = 120; // gap between upper and lower halfpipe
        public const int RectW = 20; // width

        public OnePipe(double h, double newcx)
        {
            CX = newcx;
            CY = h / 2;

            double RectH = 2 * h;
            GeometryGroup g = new GeometryGroup();
            Rect r1 = new Rect(-RectW / 2, -RectH - Gap / 2, RectW, RectH);
            Rect r2 = new Rect(-RectW / 2, Gap / 2, RectW, RectH);
            g.Children.Add(new RectangleGeometry(r1));
            g.Children.Add(new RectangleGeometry(r2));
            area = g;
        }
    }

    class FlappyModel
    {
        public const int YStep = 300; // up-down movements of pipes
        const int NumPipes = 3;

        public int Errors { get; set; }
        public OneBird Bird { get; private set; }
        public List<OnePipe> Pipes { get; private set; }
        public double GameWidth { get; private set; }
        public double GameHeight { get; private set; }

        public FlappyModel(double w, double h)
        {
            GameWidth = w;
            GameHeight = h;

            Bird = new OneBird(50, h / 2);
            Pipes = new List<OnePipe>();
            for (int i = 0; i < NumPipes; i++)
            {
                Pipes.Add(new OnePipe(h, i * w / NumPipes));
            }
        }
    }

    class FlappyLogic
    {
        static Random rnd = new Random();
        FlappyModel model;

        public FlappyLogic(FlappyModel newmodel)
        {
            model = newmodel;
        }

        public void BirdBoost(double changeSpeed)
        {
            model.Bird.DY += changeSpeed;
        }
        public void ClickScreen(Point p)
        {
            int diff = 5 * Math.Sign(p.Y - model.Bird.CY);
            BirdBoost(diff);
        }

        void BirdTick()
        {
            model.Bird.CY += model.Bird.DY;
            model.Bird.DY++;
            model.Bird.Rad = Math.Atan2(model.Bird.DY, model.Bird.DX);
        }
        void PipeTick(OnePipe pipe)
        {
            pipe.CX -= model.Bird.DX;
            if (pipe.CX < 0)
            {
                pipe.CX = model.GameWidth;
                pipe.CY = rnd.Next(
                    (int)Math.Max(pipe.CY - FlappyModel.YStep, OnePipe.Gap / 2),
                    (int)Math.Min(pipe.CY + FlappyModel.YStep, model.GameHeight - OnePipe.Gap / 2)
                );
            }
        }

        public void OneTick()
        {
            BirdTick();
            foreach (OnePipe pipe in model.Pipes)
            {
                PipeTick(pipe);
                if (model.Bird.IsCollision(pipe)) model.Errors++;
            }
        }
    }

    class FlappyRenderer
    {
        FlappyModel model;

        Pen black = new Pen(Brushes.Black, 2);
        Pen red = new Pen(Brushes.Red, 2);
        Typeface font = new Typeface("Arial");
        Point textLocation = new Point(10, 10);
        Rect bgRect;
        FormattedText formattedText;
        int oldErrorsValue = -1;

        public FlappyRenderer(FlappyModel model)
        {
            this.model = model;
            bgRect = new Rect(0, 0, model.GameWidth, model.GameHeight);
        }

        public void BuildDisplay(DrawingContext ctx)
        {
            DrawBackground(ctx);
            DrawBird(ctx);
            DrawPipes(ctx);
            DrawText(ctx);
        }

        private void DrawBackground(DrawingContext ctx)
        {
            ctx.DrawRectangle(Brushes.Gray, null, bgRect);
        }
        private void DrawBird(DrawingContext ctx)
        {
            ctx.DrawGeometry(Brushes.Red, red, model.Bird.RealArea);
        }
        private void DrawPipes(DrawingContext ctx)
        {
            foreach (OnePipe pipe in model.Pipes)
            {
                ctx.DrawGeometry(Brushes.Green, black, pipe.RealArea);
            }
        }
        private void DrawText(DrawingContext ctx)
        {
            if (oldErrorsValue != model.Errors)
            {
                formattedText = new FormattedText(
                    model.Errors.ToString(),
                    System.Globalization.CultureInfo.CurrentCulture,
                    FlowDirection.LeftToRight,
                    font,
                    16,
                    Brushes.Black, 1);
            }
            ctx.DrawText(formattedText, textLocation);
        }
    }

    class FlappyControl : FrameworkElement
    {
        FlappyModel model;
        FlappyLogic logic;
        FlappyRenderer renderer;
        DispatcherTimer mainTimer;

        public FlappyControl()
        {
            Loaded += GameScreen_Loaded;
        }

        void GameScreen_Loaded(object sender, RoutedEventArgs e)
        {
            model = new FlappyModel(ActualWidth, ActualHeight);
            logic = new FlappyLogic(model);
            renderer = new FlappyRenderer(model);

            // After GameLogic & OnRender
            Window win = Window.GetWindow(this);
            if (win != null)
            {
                win.KeyDown += GameScreen_KeyDown;
                win.MouseDown += Win_MouseDown;
                mainTimer = new DispatcherTimer();
                mainTimer.Interval = TimeSpan.FromMilliseconds(100);
                mainTimer.Tick += timer_Tick;
                mainTimer.Start();
            }

            InvalidateVisual();
        }

        protected override void OnRender(DrawingContext drawingContext)
        {
            if (renderer != null) renderer.BuildDisplay(drawingContext);
        }

        private void Win_MouseDown(object sender, MouseEventArgs e)
        {
            logic.ClickScreen(e.GetPosition(this));
        }

        // After GameLogic&OnRender
        void GameScreen_KeyDown(object sender, System.Windows.Input.KeyEventArgs e)
        {
            switch (e.Key)
            {
                case Key.Up: logic.BirdBoost(-5); break;
                case Key.Down: logic.BirdBoost(5); break;
                case Key.Space: mainTimer.IsEnabled = !mainTimer.IsEnabled; break;
            }
            InvalidateVisual();
        }
        void timer_Tick(object sender, EventArgs e)
        {
            logic.OneTick();
            InvalidateVisual();
        }
    }
}